import base64
import datetime
from functools import cached_property
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    ClassVar,
    Dict,
    List,
    Optional,
    cast,
)

import docker
from pydantic import BaseModel

if TYPE_CHECKING:
    from botocraft.services import (
        ECRImage,
        ECRImageManager,
        ImageIdentifier,
        RepositoryManager,
    )


class ECRDockerClient(BaseModel):
    """
    A return type suitable for the docker client.

    We need to return a docker client that is logged into our ECR registry,
    along with the username, password, and registry, because you need the
    latter 3 to do any pulling or pushing of images.
    """

    #: The docker client.
    client: Any
    #: The username to use for the remote registry.
    username: str
    #: The password to use for the remote registry.
    password: str
    #: The registry
    registry: str


class ImageInfo(BaseModel):
    """
    A class to hold information about a :py:class:`botocraft.services.ecr.Image`
    that is not available from the boto3 library.  We extract this information
    by pulling the image from the repository and inspecting it with the docker
    Python library.

    Important:
        You must have the docker daemon running to use the methods that return
        this object.

    """

    # The image name, including the registry, repository, and tag.
    name: str
    #: The OS platform of the image
    platform: str
    #: The architecture of the image
    architecture: str
    #: Size of the image in bytes
    size: int
    #: This is a dictionary of port mappings.  The key is the port
    #: and the value is i'm not sure what
    ports: Dict[str, Dict[str, Any]] = {}
    #: Docker Version used to build the image
    docker_version: str
    #: The user that the image runs as
    user: Optional[str] = None
    #: When the image was created, as a UTC datetime object
    created: datetime.datetime


# decorators

# Repository


def repo_list_images_ecr_images_only(
    func: Callable[..., List["ImageIdentifier"]],
) -> Callable[..., List["ECRImage"]]:
    """
    Convert a list of ECR image identifiers returned by
    :py:meth:`botocraft.services.ecr.RepositoryManager.list_images` into a list
    of :py:class:`botocraft.services.ecr.Image` objects.
    """

    def wrapper(self, *args, **kwargs) -> List["ECRImage"]:
        identifiers: List["ImageIdentifier"] = func(self, *args, **kwargs)  # noqa: UP037
        images: List["ECRImage"] = []  # noqa: UP037
        # NOTE: to be honest i'm not sure if there is a per request limit
        # for the number of images that can be retrieved, but i'm going to
        # assume that there is a limit of 100 images per request.
        for i in range(0, len(identifiers), 100):
            _images = self.get_images(
                repositoryName=args[0], imageIds=identifiers[i : i + 100]
            )
            if _images:
                images.extend(_images)
        return images

    return wrapper


# Image


def image_list_images_ecr_images_only(
    func: Callable[..., List["ImageIdentifier"]],
) -> Callable[..., List["ECRImage"]]:
    """
    Convert a list of ECR image identifiers returned by
    :py:meth:`botocraft.services.ecr.Image.list` into a list
    of :py:class:`botocraft.services.ecr.Image` objects.
    """

    def wrapper(self, *args, **kwargs) -> List["ECRImage"]:
        identifiers: List["ImageIdentifier"] = func(self, *args, **kwargs)  # noqa: UP037
        images: List["ECRImage"] = []  # noqa: UP037
        # NOTE: to be honest i'm not sure if there is a per request limit
        # for the number of images that can be retrieved, but i'm going to
        # assume that there is a limit of 100 images per request.
        for i in range(0, len(identifiers), 100):
            _images = self.get_many(
                repositoryName=args[0], imageIds=identifiers[i : i + 100]
            )
            if _images:
                images.extend(_images.images)
        return images

    return wrapper


# Mixins


class RepositoryMixin:
    objects: ClassVar["RepositoryManager"]

    # properties

    @property
    def images(self) -> Optional[List["ECRImage"]]:
        """
        Get a list of images for a given repository.
        """
        return self.objects.list_images(repositoryName=self.repositoryName)  # type: ignore[attr-defined]

    # methods

    def get_image(self, imageId: "ImageIdentifier") -> Optional["ECRImage"]:  # noqa: N803
        """
        Get an image object for a given repository and image identifier.

        Args:
            imageId: The image ID or tag to describe. The format of the imageId
                reference is ``imageTag=tag`` or ``imageDigest=digest``

        """
        return self.objects.get_image(
            self.repositoryName,  # type: ignore[attr-defined]
            imageId=imageId,
        )


class ECRImageMixin:
    """
    Add a bunch of support for inspecting ECR images and getting information
    from them that AWS does not provide.  This is done by using the docker
    Python library to pull the image and inspect it.

    Note:
        I don't love doing this because it is not pure AWS, which was my
        intention for botocraft, but I need these features for business
        purposes and they are not available in the boto3 library.

    """

    objects: ClassVar["ECRImageManager"]
    repositoryName: Optional[str]
    imageId: "ImageIdentifier"

    @property
    def version(self) -> str:
        """
        Get the version of the image.
        """
        return cast(str, self.imageId.imageTag)

    @property
    def name(self) -> str:
        """
        Get the name of the image.
        """
        return f"{self.repository.repositoryUri}:{self.imageId.imageTag}"  # type: ignore[attr-defined]

    @property
    def image_name(self) -> str:
        """
        Return just the image name, excluding the registry.
        """
        return f"{self.repositoryName}:{self.imageId.imageTag}"

    @property
    def is_pulled(self) -> bool:
        """
        Check if the image is pulled.

        Returns:
            ``True`` if the image is pulled, ``False`` otherwise.

        """
        ecr_client = self.docker_client
        exists = False
        if ecr_client.client.images.list(self.name):
            exists = True
        ecr_client.client.close()
        return exists

    @property
    def dockerd_is_running(self) -> bool:
        """
        Check if the docker daemon is running.

        We need dockerd to be running to perform these operations:

        * :py:meth:`docker_client`
        * :py:meth:`pull`
        * :py:meth:`is_pulled`
        * :py:meth:`info`
        * :py:meth:`docker_image`
        * :py:meth:`history`
        * :py:meth:`clean`
        * :py:meth:`clean_other_versions`
        """
        try:
            docker.from_env()
        except docker.errors.DockerException:
            return False
        return True

    @property
    def docker_client(self) -> ECRDockerClient:
        """
        Return a docker client, logged into our ECR registry.

        Raises:
            RuntimeError: If the docker daemon is not running.

        Returns:
            A :py:class:`botocraft.mixins.ecr.ECRDockerClient` object, which
            has a docker client, username, password, and registry.

        """
        if not self.dockerd_is_running:
            msg = "Docker daemon is not running, so this command is not available."
            raise RuntimeError(msg)
        docker_client = docker.from_env()
        # Get our authorization token from AWS
        response = self.objects.using(self.session).client.get_authorization_token()  # type: ignore[attr-defined]
        auth_token = base64.b64decode(
            response["authorizationData"][0]["authorizationToken"]
        )
        username, password = auth_token.decode().split(":")
        registry = response["authorizationData"][0]["proxyEndpoint"]
        bare_registry = registry.split("//")[1]
        docker_client.login(username, password=password, registry=registry, reauth=True)
        return ECRDockerClient(
            client=docker_client,
            username=username,
            password=password,
            registry=bare_registry,
        )

    @property
    def info(self) -> ImageInfo:
        """
        Return information about the image.  We're doing this by pulling the
        image from the repository and inspecting it.

        Note:
            I'd love to get the base image for this image, but there is no
            direct way to do it.  You would to look up the layers for the image,
            get the sha256 hash of the first layer (which is the base image),
            then look in in various repositories to find the image that
            has the same layer, then get that image's name.  That seems stupid
            hard to do, especially if the base image is in the ECR registry of
            another AWS account.

        Raises:
            RuntimeError: If the docker daemon is not running.

        Returns:
            A :py:class:`botocraft.services.ecr.ImageInfo` object.

        """
        ecr_client = self.docker_client
        data = ecr_client.client.api.inspect_image(self.name)
        # you can't be logged into two ECR registries at the same time for some reason
        # so we need to log out of the registry we are using.
        ecr_client.client.close()
        # Strip off the nanoseconds from the created date so that strptime can
        # parse it.
        created_date = data["Created"].split(".")[0] + "Z"
        return ImageInfo(
            name=data["RepoTags"][0],
            platform=data["Os"],
            architecture=data["Architecture"],
            size=data["Size"],
            docker_version=data["DockerVersion"],
            user=data["Config"]["User"],
            ports=data["Config"]["ExposedPorts"],
            # Created date looks like: '2024-08-19T21:59:57', convert
            # that to a datetime object.
            created=datetime.datetime.strptime(created_date, "%Y-%m-%dT%H:%M:%SZ"),  # noqa: DTZ007
        )

    @cached_property
    def docker_image(self) -> docker.models.images.Image:
        """
        Return the :py:class:`docker.models.images.Image` object for this image.

        Raises:
            RuntimeError: If the docker daemon is not running.

        """
        ecr_client = self.docker_client
        if not self.is_pulled:
            docker_image = ecr_client.client.images.pull(
                f"{ecr_client.registry}/{self.repositoryName}",
                auth_config={
                    "username": ecr_client.username,
                    "password": ecr_client.password,
                },
                tag=self.imageId.imageTag,
            )
        else:
            docker_image = ecr_client.client.images.get(self.name)
        ecr_client.client.close()
        return docker_image

    @cached_property
    def history(self) -> List[Dict[str, Any]]:
        """
        Return the build history for this image.  You can use this to reconstruct
        **most** of the Dockerfile that was used to build the image.   You won't
        have the ``FROM`` line, but you can get most of the rest of it.

        Raises:
            RuntimeError: If the docker daemon is not running.

        """
        return self.docker_image.history()

    # method
    def clean(self) -> None:
        """
        Remove the image from our local docker storage, if it exists.

        Raises:
            RuntimeError: If the docker daemon is not running.

        """
        if self.is_pulled:
            ecr_client = self.docker_client
            ecr_client.client.images.remove(self.name)
            ecr_client.client.close()

    # method
    def clean_other_versions(self) -> None:
        """
        Remove the all images for this repository except for the one with
        our version.

        Raises:
            RuntimeError: If the docker daemon is not running.

        """
        ecr_client = self.docker_client
        prefix = f"{ecr_client.registry}/{self.repositoryName}"
        images = ecr_client.client.images.list(prefix)
        for image in images:
            if self.name not in image.tags:
                ecr_client.client.images.remove(f"{prefix}:{image.imageTag}")
        ecr_client.client.close()



import re
from collections import OrderedDict
from typing import cast

from botocraft.sync.models import MethodArgumentDefinition

from ...models import MethodDocstringDefinition
from .base import ManagerMethodGenerator


class CreateMethodGenerator(ManagerMethodGenerator):
    """
    Handle the generation of a create method.
    """

    method_name: str = "create"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.method_def.args.update(self.get_explicit_args_from_request())

    @property
    def signature(self) -> str:
        """
        For create methods, we add the model as the only argument.

        Returns:
            The method signature for the method.

        """
        if self.model_def.alternate_name:
            model_name = self.model_def.alternate_name
        else:
            model_name = self.model_name
        signature = f'    def {self.method_name}(self, model: "{model_name}"'
        if self.explicit_args or self.explicit_kwargs:
            signature += ", "
        signature += ", ".join(
            [f"{arg}: {arg_type}" for arg, arg_type in self.explicit_args.items()]
        )
        if self.explicit_args and self.explicit_kwargs:
            signature += ", "
        signature += ", ".join(
            [
                f"{arg}: {arg_type} = {self.method_def.args.get(arg, MethodArgumentDefinition()).default}"  # noqa: E501
                for arg, arg_type in self.explicit_kwargs.items()
            ]
        )
        signature += f") -> {self.return_type}:"
        return signature

    @property
    def operation_args(self) -> str:
        """
        Get positional argument string for the boto3 operation call.

        Example:

            .. code-block:: python

                "arg1=data['arg1'], arg2=data['arg2']"

        We're overriding this because we need to map get most of our arguments
        for the boto3 call from the model attributes, not the method arguments,
        which is what
        :py:meth:`botocraft.sync.methods.base.MethodGenerator.operation_args`
        provides.

        Returns:
            The string to use to represent the boto3 operation positional
            arguments.

        """
        mapping = self.method_def.args
        _args: OrderedDict[str, str] = OrderedDict()
        for arg in self.args(location="operation"):
            arg_def = mapping.get(arg, MethodArgumentDefinition())
            if arg in self.explicit_args:
                _arg = self.serialize(arg)
            else:
                if arg_def.hidden:
                    continue
                _arg = f"data.get('{arg}')"
                if arg_def.value:
                    _arg = self.serialize(cast(str, arg_def.value))
                elif arg_def.source_arg:
                    _arg = self.serialize(cast(str, arg_def.source_arg))
                elif arg_def.attribute:
                    _arg = f"data['{arg_def.attribute}']"
            _args[arg] = _arg
        return ", ".join([f"{arg}={_arg}" for arg, _arg in _args.items()])

    @property
    def operation_kwargs(self) -> str:
        """
        Get keyword argument string for the boto3 operation call.

        Example:

            .. code-block:: python

                "arg1=data['arg1'], arg2=data['arg2']"

        We're overriding this because we need to map get most of our arguments
        for the boto3 call from the model attributes, not the method arguments,
        which is what
        :py:meth:`botocraft.sync.methods.base.MethodGenerator.operation_kwargs`
        provides.

        Returns:
            The string to use to represent the boto3 operation positional
            arguments.

        """
        mapping = self.method_def.args
        _args = OrderedDict()
        for arg in self.kwargs(location="operation"):
            arg_def = mapping.get(arg, MethodArgumentDefinition())
            if arg in self.explicit_kwargs:
                _arg = self.serialize(arg)
            else:
                if arg_def.hidden:
                    continue
                _arg = f"data.get('{arg}')"
                if arg_def.value:
                    if arg_def.raw_value:
                        _arg = arg_def.value
                    else:
                        _arg = self.serialize(cast(str, arg_def.value))
                elif arg_def.source_arg:
                    _arg = self.serialize(cast(str, arg_def.source_arg))
                elif arg_def.attribute:
                    _arg = f"data['{arg_def.attribute}']"
            _args[arg] = _arg
        return ", ".join([f"{arg}={_arg}" for arg, _arg in _args.items()])

    @property
    def operation_call(self) -> str:
        """
        A body snippet that does the actual boto3 call.  and assigns the
        response to ``_response``, then loads the ``_response`` into
        :py:meth:`response_class``.

        This is different from the base class, because we need to do a few things:

        * We need to map model attributes to the boto3 call arguments.
        * In some cases the boto3 call arguments are different from the model
          attributes, so we need to override the mapping with something
        * Some boto3 call arguments are not actually model attributes, so we
          need to explicitly pass them in.

        It then uses response to instantiate :py:meth:`response_class`.

        Returns:
            The code for the boto3 call.

        """
        call = "data = model.model_dump(exclude_none=True, by_alias=True)"
        args = self.operation_args
        kwargs = self.operation_kwargs
        if args or kwargs:
            call += "\n        args = dict("
        if args:
            call += args
        if args and kwargs:
            call += ", "
        if kwargs:
            call += kwargs
        call += ")"
        if self.return_type == "None":
            call += f"\n        self.client.{self.boto3_name}(**{{k: v for k, v in args.items() if v is not None}})"  # noqa: E501
        else:
            call += f"\n        _response = self.client.{self.boto3_name}(**{{k: v for k, v in args.items() if v is not None}})"  # noqa: E501
            call += f"\n        response = {self.response_class}(**_response)"
        return call

    @property
    def docstrings_def(self) -> MethodDocstringDefinition:
        """
        Return the docstring for the method.
        """
        docstrings: MethodDocstringDefinition = MethodDocstringDefinition()
        docstrings.method = (
            self.method_def.docstring
            if self.method_def.docstring
            else self.operation_model.documentation
        )
        docstrings.args["model"] = f"The :py:class:`{self.model_name}` to create."
        for arg in self.explicit_args:
            docstrings.args[arg] = self.get_arg_docstring(arg)
        for arg in self.explicit_kwargs:
            docstrings.kwargs[arg] = self.get_arg_docstring(arg)
        return docstrings

    @property
    def body(self) -> str:
        response_attr = cast(str, self.response_attr)
        if self.response_attr_multiplicity == "many":
            response_attr = f"{response_attr}[0]"
        code = f"""
        {self.operation_call}
"""
        if self.return_type != "None":
            code += f"""
        self.sessionize(response.{response_attr})
        return cast({self.return_type}, response.{response_attr})
"""
        return code

    @property
    def return_type(self) -> str:
        """
        For create, update and delete methods, we return the model itself, not
        the response model, unless it's overriden in our botocraft method
        config, in which case we return that.

        Returns:
            The name of the return type class.

        """
        return_type = super().return_type
        # Sometimes the return type is a list of the model, e.g.
        # ``elbv2:create_load_balancer``, so we need to strip the ``List`` part of the
        # return type, because we want to return just the model, if possible.
        if return_type.startswith("List["):
            return_type = re.sub(r"List\[(.*)\]", r"\1", return_type)
        return return_type



import re
from collections import OrderedDict
from textwrap import (
    indent as add_prefix,
)
from textwrap import (
    wrap,
)
from typing import Dict, List, Optional

import botocore.session
import click

from .cli import cli


def camel_to_snake(camel_str: str) -> str:
    """
    Convert a camel case string to snake case.

    Args:
        camel_str: the input camel case string

    Returns:
        The snake case version of the input string

    """
    return re.sub(r"(?<=[a-z])(?=[A-Z])", "_", camel_str).lower()


def print_shape(
    service_model: botocore.model.ServiceModel,
    shape_name: str,
    indent: int = 0,
    label: Optional[str] = None,
    documentation: bool = False,
) -> None:
    """
    Print the name and members of a shape.

    Args:
        service_model: the botocore service model object
        shape_name: the name of the shape to print

    Keyword Args:
        indent: the number of spaces to indent the output
        label: a label to print before the shape name
        documentation: whether to print the shape documentation

    """
    shape = service_model._shape_resolver.get_shape_by_name(shape_name)  # type: ignore[attr-defined]  # noqa: SLF001
    output = []
    if shape.type_name == "structure":
        if label is not None:
            output.append(
                f"{click.style(label, fg='cyan')}: {click.style(shape_name, fg='red')}"
            )
        else:
            output.append(click.style(f"{shape_name}:", fg="red"))
        if documentation and hasattr(shape, "documentation") and shape.documentation:
            docs = wrap(shape.documentation)
            output.append(f"    {click.style(docs, fg='green')}")
        if hasattr(shape, "members") and shape.members:
            for member_name, member_shape in shape.members.items():
                output.append(
                    f"    {click.style(member_name, fg='cyan')}: {member_shape.type_name} -> "  # noqa: E501
                    f"{click.style(member_shape.name, fg='blue')}"
                )
        else:
            output.append("    No members")
    elif shape.type_name == "list":
        output.append(
            f"{click.style(shape_name, fg='red')}: List -> {click.style(shape.member.name, fg='blue')}"  # noqa: E501
        )
    elif shape.type_name == "string":
        output.append(
            f"{click.style(shape_name, fg='red')}: String -> {click.style(shape.name, fg='blue')}"  # noqa: E501
        )
        if shape.enum:
            values = ", ".join(shape.enum)
            output.append(f"    Enum: {click.style(values, fg='blue')}")
    # purge empty lines
    output = [line for line in output if line.strip()]
    _output = "\n".join(output)
    if indent:
        _output = add_prefix(_output, " " * indent)
    print(_output)


@cli.group(short_help="Inspect botocore definitions", name="botocore")
def botocore_group():
    pass


@botocore_group.command("services", short_help="List all available botocore services")
def botocore_list_services():
    """
    List codenames and human names for all botocore services.
    """
    session = botocore.session.get_session()
    for service_name in session.get_available_services():
        service_model = session.get_service_model(service_name)
        print(
            f"{click.style(service_name, fg='blue')}: "
            f"{service_model.metadata['serviceId']}"
        )


@botocore_group.command("models", short_help="List all available shapes for a service")
@click.option("--names-only", is_flag=True, help="List only model names, not shapes")
@click.argument("service")
def botocore_list_shapes(service: str, names_only: bool):
    """
    List all shapes in a botocore service model.

    Args:
        service: the name of the service
        names_only: whether to list only the names of the shapes

    """
    session = botocore.session.get_session()
    service_model = session.get_service_model(service)
    for shape_name in service_model.shape_names:  # pylint: disable=not-an-iterable
        if names_only:
            click.secho(shape_name, fg="red")
        else:
            print_shape(service_model, shape_name)


@botocore_group.command("model", short_help="List all available shapes for a service")
@click.option("--dependencies", is_flag=True, help="List dependencies for the model")
@click.option("--operations", is_flag=True, help="List operations for the model")
@click.option("--documentation", is_flag=True, help="Show documentation for the model")
@click.argument("service")
@click.argument("model")
def botocore_list_shape(
    service: str,
    model: str,
    dependencies: bool,
    operations: bool,
    documentation: bool,
):
    session = botocore.session.get_session()
    service_model = session.get_service_model(service)
    if model not in list(service_model.shape_names):
        click.secho(f"Model {model} not found in service {service}", fg="red")
    print_shape(service_model, model, documentation=documentation)
    if operations:
        _operations = [op for op in list(service_model.operation_names) if model in op]
        if _operations:
            print()
            click.secho("Operations:", fg="yellow")
            click.secho("-" * len("Operations"), fg="yellow")
            print()
            for operation in _operations:
                click.secho(camel_to_snake(operation), fg="cyan")
    if dependencies:
        print()
        click.secho("Dependencies:", fg="yellow")
        click.secho("-" * len("Dependencies"), fg="yellow")
        print()
        shape = service_model._shape_resolver.get_shape_by_name(model)  # type: ignore[attr-defined]  # noqa: SLF001
        if (
            shape.type_name == "structure"
            and hasattr(shape, "members")
            and shape.members
        ):
            for member_name, member_shape in shape.members.items():
                if member_shape.type_name == "structure":
                    click.secho(f"{model}.{member_name}:", fg="cyan")
                    print_shape(service_model, member_shape.name, indent=4)
                elif member_shape.type_name == "list":
                    list_shape = member_shape.member
                    click.secho(f"{model}.{member_name} -> List:", fg="cyan")
                    if list_shape.type_name == "structure":
                        print_shape(service_model, list_shape.name, indent=4)
                elif member_shape.type_name == "string":
                    if member_shape.enum:
                        click.secho(f"{model}.{member_name}:", fg="cyan")
                        click.secho(f"    {member_name} -> Enum:", fg="cyan")
                        values = ", ".join(member_shape.enum)
                        click.secho(f"      {values}", fg="white")


@botocore_group.command(
    "operations", short_help="List all available operations for a service"
)
@click.argument("service")
def botocore_list_operations(service: str):
    """
    Print all operations for a service, along with their input and output shapes.

    Args:
        service: the name of the service

    """
    session = botocore.session.get_session()
    service_model = session.get_service_model(service)
    for name in service_model.operation_names:  # pylint: disable=not-an-iterable
        operation_model = service_model.operation_model(name)
        print(f"{name}:")
        boto3_name = camel_to_snake(name)
        print(f"    boto3 name: {boto3_name}")
        input_shape = operation_model.input_shape
        if input_shape is not None:
            print_shape(service_model, input_shape.name, indent=4, label="Input")
        output_shape = operation_model.output_shape
        if output_shape is not None:
            print_shape(service_model, output_shape.name, indent=4, label="Output")


@botocore_group.command(
    "primary-models", short_help="List all probable primary models for a service"
)
@click.argument("service")
def botocore_list_primary_models(service: str):
    """
    List all probable primary models for a service.

    Args:
        service: the name of the service

    """
    session = botocore.session.get_session()
    service_model = session.get_service_model(service)
    operation_names: List[str] = list(service_model.operation_names)
    prefixes = (
        "Put",
        "Get",
        "Create",
        "Delete",
        "Describe",
        "List",
        "Update",
        "Modify",
    )
    writable_prefixes = ("Put", "Create", "Delete", "Update", "Modify")
    # First pass: list all shapes
    # Second pass: assign operations to the most specific shape
    # Then print the shapes with their operations
    models: Dict[str, List[str]] = {}
    names = list(service_model.shape_names)
    names.sort(key=lambda x: len(x))
    names.reverse()
    taken = []
    for shape_name in names:
        shape = service_model._shape_resolver.get_shape_by_name(shape_name)  # type: ignore[attr-defined]  # noqa: SLF001
        if shape.type_name != "structure":
            continue
        operations = [
            op
            for op in operation_names
            if shape_name in op and op.startswith(prefixes) and op not in taken
        ]
        if operations:
            models[shape_name] = operations
            taken.extend(operations)
    _models = OrderedDict(sorted(models.items(), key=lambda x: x[1]))
    for model in _models:
        operations = _models[model]
        writable: bool = False
        label: str = ""
        for op in operations:
            if op.startswith(writable_prefixes):
                writable = True
                break
        if not writable:
            label = click.style(": [READONLY]", fg="green")
        click.echo(f'{click.style(model, fg="red")}{label}')
        for operation in operations:
            click.secho(f"    {camel_to_snake(operation)}", fg="cyan")



import re
from typing import cast

from .base import ManagerMethodGenerator


class PartialUpdateMethodGenerator(ManagerMethodGenerator):
    """
    Handle the generation of a partial_update method.

    This is a special case of the update method, where we only update a subset
    of the model attributes.   In this case, we don't pass in a model instance,
    but just a pk and the attributes to update.  This allows us to do simple
    updates like ``scale`` without having to fetch the model first.
    """

    method_name: str = "partial_update"

    @property
    def body(self) -> str:
        response_attr = None
        if self.response_attr is not None:
            response_attr = cast(str, self.response_attr)
            if self.response_attr_multiplicity == "many":
                response_attr = f"{response_attr}[0]"
        code = f"""
        {self.operation_args}
        {self.operation_call}
"""
        if response_attr is not None:
            code += f"""
        self.sessionize(response.{response_attr})
        return cast({self.return_type}, response.{response_attr})
"""
        return code

    @property
    def return_type(self) -> str:
        """
        We want to return a single model instance, not a list of them, so we
        override the return type here.

        Returns:
            The name of the return type class.

        """
        return_type = super().return_type
        # Sometimes the return type is a list of the model, e.g.
        # ``elbv2:create_load_balancer``, so we need to strip the ``List`` part of the
        # return type, because we want to return just the model, if possible.
        if return_type.startswith("List["):
            return_type = re.sub(r"List\[(.*)\]", r"\1", return_type)
        return return_type



from typing import Optional

from .base import ManagerMethodGenerator


class DeleteMethodGenerator(ManagerMethodGenerator):
    """
    Generate the code for the delete method.

    boto3 delete methods typically return the dict of the full object that was
    deleted, so we'll convert that to the model we're working with and return
    that instead.
    """

    method_name: str = "delete"

    @property
    def response_attr(self) -> Optional[str]:
        """
        Override our superclass here to not throw an error if we can't find the
        response attribute.  This is because sometimes delete methods return
        nothing, so we'll just return None.

        Returns:
            The response attribute name, or ``None``.

        """
        try:
            return super().response_attr
        except ValueError:
            return None

    @property
    def return_type(self) -> str:
        """
        For delete methods, we return the model itself, not the response model,
        unless it's overriden in our botocraft method config, in which case
        we return that.

        Returns:
            The name of the return type class.

        """
        # Leave this here because the response_class property creates the
        # response class code if it doesn't exist.
        _ = self.response_class
        return_type = super().return_type
        if return_type != "None":
            return_type = f'"{self.model_name}"'
            if self.method_def.return_type:
                return_type = self.method_def.return_type
        return return_type

    @property
    def body(self) -> str:
        """
        Generate the method body code for the delete method.
        """
        code = f"""
        {self.operation_args}
        {self.operation_call}
"""
        if self.return_type != "None":
            if self.response_attr is None:
                code += "        return response"
            else:
                code += f"        return cast({self.model_name}, response.{self.response_attr})"  # noqa: E501
        return code



import importlib.util
import subprocess
import sys

from .cli import cli


@cli.command(
    short_help="Run an interactive python shell with all services loaded", name="shell"
)
def shell_command():
    """
    Start an interactive Python shell with preloaded imports.
    """
    # Import botocraft.services
    preload_command = "from botocraft.services import *; import boto3"

    # Check if IPython is installed
    ipython_installed = importlib.util.find_spec("IPython") is not None

    if ipython_installed:
        # If IPython is installed, start IPython with the preloaded command
        subprocess.run(["ipython", "-i", "-c", preload_command], check=False)
    else:
        # If IPython is not installed, start the default Python shell with the
        # preloaded command
        subprocess.run([sys.executable, "-i", "-c", preload_command], check=False)



import re
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from botocraft.sync.service import ModelGenerator


class ModelPropertyGenerator:
    """
    Base class for all model method properties.  This is used to generate
    the code for a single method on a manager class.

    To use this, you subclass it and implement the :py:meth:`body`
    property, which is the body of the method.  You can also override
    any of the following properties to customize the method:

    * :py:meth:`decorator`: The method decorators.
    * :py:meth:`signature`: The method signature.
    * :py:meth:`docstring`: The method docstring.
    * :py:meth:`return_type`: The return type annotation.

    Args:
        generator: The generator that is creating the model class
        model_name: The name of the model we're generating the property for.

    """

    def __init__(
        self,
        generator: "ModelGenerator",
        model_name: str,
        property_name: str,
    ) -> None:
        #: The generator that is creating the service classes for an AWS Service.
        self.generator = generator
        #: The name of the model we're generating the property for.
        self.model_name = model_name
        #: The definition of the model we're generating the property for.
        self.model_def = self.generator.get_model_def(model_name)
        #: The name of the property we're generating.
        self.property_name = property_name
        #: The definition of the property we're generating.
        self.property_def = self.model_def.properties[self.property_name]

    @property
    def decorator(self) -> str:
        """
        The decorator for the method.  If the property definition has
        :py:attr:`botocraft.sync.models.ModelPropertyDefinition.cached`` equal
        to ``True``, this is ``@cached_property``, otherwise use ``@property``.

        Returns:
            The decorator for the method.

        """
        if self.property_def.cached:
            self.generator.imports.add("from functools import cached_property")
            return "    @cached_property"
        return "    @property"

    @property
    def return_type(self) -> str:
        """
        Return the return type annotation for the method.
        """
        if self.property_def.transformer.regex:
            return_type = "Optional[str]"
            num_groups = re.compile(self.property_def.transformer.regex.regex).groups
            if num_groups > 1:
                return_type = "Optional[Dict[str, str]]"
        elif self.property_def.transformer.mapping:
            # TODO: it'd be nice to do something like a typed dict here
            return_type = "OrderedDict[str, Any]"
            self.generator.imports.add("from collections import OrderedDict")
        elif self.property_def.transformer.alias:
            fields = self.generator.fields(self.model_name)
            assert (
                self.property_def.transformer.alias in fields
            ), f"Alias: attribute {self.property_def.transformer.alias} not found in model {self.model_name}"  # noqa: E501
            return_type = self.generator.field_type(
                self.model_name,
                self.property_def.transformer.alias,
                field_shape=fields[self.property_def.transformer.alias].botocore_shape,
            )
        return return_type

    @property
    def docstring(self) -> str:
        """
        Return the docstring for the method.
        """
        if self.property_def.docstring:
            return f'''
        """
        {self.property_def.docstring}

        """
'''
        return ""

    @property
    def signature(self) -> str:
        """
        Return the method signature.

        Returns:
            The method signature.

        """
        return f"    def {self.property_name}(self) -> {self.return_type}:"

    @property
    def _regex_body(self) -> str:
        """
        Render the body of the method for use with the regex transformer.

        Returns:
            The method body.

        """
        assert (
            self.property_def.transformer.regex
        ), f"Property {self.property_name} does not have a regex transformer"
        return f"""
        return self.transform(
            "{self.property_def.transformer.regex.attribute}",
            r"{self.property_def.transformer.regex.regex}"
        )
"""

    @property
    def _mapping_body(self) -> str:
        """
        Render the body of the method for use with the mapping transformer.

        Returns:
            The method body.

        """
        assert (
            self.property_def.transformer.mapping
        ), f"Property {self.property_name} does not have a mapping transformer"
        code = """        return OrderedDict({
"""
        for key, value in self.property_def.transformer.mapping.items():
            code += f"""
            "{key}": self.{value},
"""
        code += """
        })
"""
        return code

    @property
    def _alias_body(self) -> str:
        """
        Render the body of the method for use with the alias transformer.

        Returns:
            The method body.

        """
        assert (
            self.property_def.transformer.alias
        ), f"Property {self.property_name} does not have an alias transformer"
        return f"""
        return self.{self.property_def.transformer.alias}
"""

    @property
    def body(self) -> str:
        """
        Return the method body.

        Returns:
            The method body.

        """
        if self.property_def.transformer.regex:
            return self._regex_body
        if self.property_def.transformer.mapping:
            return self._mapping_body
        if self.property_def.transformer.alias:
            return self._alias_body
        msg = f"Property {self.property_name} does not have a transformer"
        raise RuntimeError(msg)

    @property
    def code(self) -> str:
        """
        Return the code for the method.

        Returns:
            The code for the method.

        """
        return f"""

{self.decorator}
{self.signature}
{self.docstring}
{self.body}
"""
